/*
 * This file is part of Jstacs.
 * 
 * Jstacs is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 * 
 * Jstacs is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * Jstacs. If not, see <http://www.gnu.org/licenses/>.
 * 
 * For more information on Jstacs, visit http://www.jstacs.de
 */

package de.jstacs.io;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.Date;

/**
 * This class is for handling {@link File}s. The most important methods of this
 * class are for writing a {@link StringBuffer} to or reading it from a
 * {@link File}. This is useful for all objects that implement
 * {@link de.jstacs.Storable} and should be saved in or loaded from a
 * {@link File}.
 * 
 * @author Jens Keilwagen
 * 
 * @see de.jstacs.Storable
 * @see File
 */
public class FileManager {

	private FileManager() {}

	/**
	 * This buffer is used to copy {@link File}s.
	 */
	private static byte[] buffer = new byte[100000];

	/**
	 * This method copies all {@link File}s and directories, if selected, from a
	 * <code>source</code> {@link File}, i.e. directory, to a
	 * <code>target</code> {@link File}, i.e. directory, that have been modified
	 * after a predefined <code>date</code>.
	 * 
	 * @param source
	 *            the source directory denoted as {@link File}
	 * @param target
	 *            the target directory denoted as {@link File}
	 * @param date
	 *            the {@link Date}
	 * @param recursive
	 *            indicates if this method should be used recursively in the
	 *            subdirectories
	 * @param filter
	 *            a {@link FileFilter} for the {@link File}s that enables the
	 *            user to copy only specific {@link File}s
	 * 
	 * @return the number of copied {@link File}s
	 * 
	 * @throws IllegalArgumentException
	 *             if <code>source</code> and <code>target</code> are not
	 *             directories
	 * @throws IOException
	 *             if something went wrong while copying the {@link File}s and
	 *             directories
	 * 
	 * @see File
	 * @see Date
	 * @see FileFilter
	 */
	public static int copyDiff( File source, File target, Date date, boolean recursive, FileFilter filter ) throws IllegalArgumentException,
			IOException {
		if( !source.isDirectory() || !target.isDirectory() ) {
			throw new IllegalArgumentException( "The source and the target have to be directories." );
		}
		File[] files = source.listFiles( filter );
		//System.out.println( "xxx " + files.length );
		int anz = 0, i;
		long time = date.getTime();
		boolean existsBefore;
		File help;
		for( File f : files ) {
			if( f.isDirectory() ) {
				if( recursive ) {
					help = new File( target.getAbsolutePath() + "/" + f.getName() );
					existsBefore = help.exists();
					if( !existsBefore ) {
						help.mkdir();
					}
					i = copyDiff( f, help, date, recursive, filter );
					if( i == 0 ) {
						if( !existsBefore ) {
							help.delete();
						}
					} else {
						anz += i;
					}
				}
			} else {
				if( f.lastModified() >= time ) {
					//System.out.println( f.getName() + "\t" + target.getAbsolutePath() + "/" + f.getName() );

					copy( f.getAbsolutePath(), target.getAbsolutePath() + "/" + f.getName() );
					anz++;
				}
			}
		}
		return anz;
	}

	/**
	 * This method copies a {@link File} in a faster manner.
	 * 
	 * @param from
	 *            the {@link File} name of the original file
	 * @param to
	 *            the {@link File} name of the copied file
	 * 
	 * @throws IOException
	 *             if something went wrong
	 * 
	 * @see FileManager#copy(String, String, byte[])
	 */
	public static void copy( String from, String to ) throws IOException {
		copy( from, to, buffer );
	}

	/**
	 * This method copies a {@link File} in a faster manner using a specified
	 * buffer.
	 * 
	 * @param from
	 *            the {@link File} name of the original file
	 * @param to
	 *            the {@link File} name of the copied file
	 * @param buffer
	 *            an array for reading the content of the original {@link File},
	 *            the size of the array determines how many <code>byte</code>s
	 *            are read at once
	 * 
	 * @throws IOException
	 *             if something went wrong
	 */
	public static synchronized void copy( String from, String to, byte[] buffer ) throws IOException {
		FileInputStream in = null;
		FileOutputStream out = null;
		try {
			in = new FileInputStream( from );
			out = new FileOutputStream( to );
			int amountRead;
			while( ( amountRead = in.read( buffer ) ) > -1 ) {
				out.write( buffer, 0, amountRead );
			}
		} finally {
			if( in != null ) {
				in.close();
			}
			if( out != null ) {
				out.close();
			}
		}
	}

	/**
	 * This method reads a {@link StringBuffer} from a given {@link File}.
	 * 
	 * @param file
	 *            the {@link File} to be read
	 * 
	 * @return a {@link StringBuffer} with the content of the {@link File}
	 * 
	 * @throws IOException
	 *             if something went wrong with the {@link File}
	 * 
	 * @see FileManager#writeFile(File, StringBuffer)
	 * @see FileManager#read(Reader)
	 * @see FileReader
	 */
	public static StringBuffer readFile( File file ) throws IOException {
		return read( new FileReader( file ) );
	}

	/**
	 * This method reads a {@link StringBuffer} from a given {@link InputStream}.
	 * 
	 * @param inputStream
	 *            the {@link InputStream} to be read
	 * 
	 * @return a {@link StringBuffer} with the content of the {@link InputStream}
	 * 
	 * @throws IOException
	 *             if something went wrong with the {@link InputStream}
	 * 
	 * @see FileManager#writeFile(File, StringBuffer)
	 * @see FileManager#read(Reader)
	 * @see InputStreamReader
	 */
	public static StringBuffer readInputStream( InputStream inputStream ) throws IOException {
		return read( new InputStreamReader( inputStream ) );
	}
	
	/**
	 * This method reads a {@link StringBuffer} from a given {@link Reader}.
	 * 
	 * @param reader
	 *            the {@link Reader} to be read
	 * 
	 * @return a {@link StringBuffer} with the content of the {@link Reader}
	 * 
	 * @throws IOException
	 *             if something went wrong with the {@link Reader}
	 * 
	 * @see FileManager#writeFile(File, StringBuffer)
	 * @see BufferedReader
	 */
	public static StringBuffer read( Reader reader ) throws IOException {
		BufferedReader r = new BufferedReader( reader, 100000 );
		StringBuffer res = new StringBuffer( 1000000 );
		String help;
		while( ( help = r.readLine() ) != null ) {
			res.append( help + "\n" );
		}
		r.close();
		return res;
	}

	/**
	 * This method saves a {@link StringBuffer} to a given {@link File}.
	 * 
	 * @param outputFile
	 *            the {@link File} into which the output should be written
	 * @param buffer
	 *            the buffer to be written in the {@link File}
	 * 
	 * @throws IOException
	 *             if something went wrong with the {@link File}
	 * 
	 * @see FileManager#readFile(File)
	 */
	public static void writeFile( File outputFile, StringBuffer buffer ) throws IOException {
		BufferedWriter b = new BufferedWriter( new FileWriter( outputFile ) );
		b.write( buffer.toString() );
		b.close();
	}
}
